import sys
import math
import itertools
import collections
import random
import pygame
import ptext

import data
import const
import util

class GameWindow(object):
    
    def __init__(self, fullscreen=False):
        pygame.display.set_caption('Battle Critters')
        self.screen = pygame.display.set_mode((const.WIDTH, const.HEIGHT))
        self.intro()

    def intro(self):
        Intro(self, self.screen).loop()
        Intro(self, self.screen, 'howtoplay').loop()
        self.game()
    
    def game(self):
        restart = True
        while restart:
            game = Game(self, self.screen)
            restart = game.loop()        


class Intro(object):

    def __init__(self, window, screen, image='intro'):
        self.window = window
        self.real_screen = screen
        self.screen = pygame.surface.Surface((const.REAL_WIDTH, const.REAL_HEIGHT))
        self.bg = pygame.image.load(data.filepath(image + ".png")).convert()
        self.screen.blit(self.bg, (0, 0))

        pygame.transform.scale(self.screen, (const.WIDTH, const.HEIGHT), self.real_screen)
        pygame.display.update()
        
    def loop(self):
        self.done = False
        while not self.done:
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    sys.exit()
                elif event.type == pygame.KEYDOWN:
                    if event.key == pygame.K_ESCAPE:
                        sys.exit()
                    else:
                        self.done = True
                elif event.type == pygame.MOUSEBUTTONDOWN:
                    self.done = True


class Game(object):

    def __init__(self, window, screen):
        self.window = window
        self.real_screen = screen
        self.screen = pygame.surface.Surface((const.REAL_WIDTH, const.REAL_HEIGHT))
        self.clock = pygame.time.Clock()

        try:
            pygame.mixer.init()
            pygame.mixer.music.load(data.filepath("BrittleRille.ogg"))
            pygame.mixer.music.play(-1)
            pygame.mixer.music.set_volume(0.1)
        except:
            print 'Cannot load sound. :('

        self.slash_mgr = SlashManager()
        self.critter_mgr = CritterManager(self.slash_mgr)
        self.tile_mgr = TileManager()
        self.board_mgr = BoardManager(self.critter_mgr, self.tile_mgr)
        self.board_mgr.setup()

        self.paused = False

        self.fullscreen = False
        self.just_changed_fullscreen = False

    def loop(self):
        self.bg = pygame.surface.Surface((const.WIDTH, const.HEIGHT)).convert_alpha()
        self.screen.blit(self.bg, (0, 0))

        pygame.transform.scale(self.screen, (const.WIDTH, const.HEIGHT), self.real_screen)
        pygame.display.update()

        self.done = False

        while not self.done:
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    sys.exit()
                elif event.type == pygame.KEYDOWN:
                    self.on_keydown(event)
                elif event.type == pygame.MOUSEBUTTONDOWN:
                    self.on_mousedown(event)

            if not self.paused:
                self.tile_mgr.clear(self.screen, self.bg)
                self.critter_mgr.clear(self.screen, self.bg)
                self.slash_mgr.clear(self.screen, self.bg)
                self.board_mgr.clear(self.screen, self.bg)

            dt = self.clock.tick(const.FPS) / 1000.0

            if not self.paused:
                self.board_mgr.update(dt)
                self.tile_mgr.update(dt)
                self.critter_mgr.update(dt)
                self.slash_mgr.update(dt)

            update_rects = []
            if not self.paused:
                update_rects += self.tile_mgr.draw(self.screen)
                update_rects += self.critter_mgr.draw(self.screen)
                update_rects += self.slash_mgr.draw(self.screen)
                update_rects += self.board_mgr.draw(self.screen)

            scaled_update_rects = [pygame.Rect(2*r.left - 4, 2*r.top - 4, 2*r.width + 8, 2*r.height + 8) for r in update_rects]

            pygame.transform.scale(self.screen, (const.WIDTH, const.HEIGHT), self.real_screen)
            pygame.display.update(scaled_update_rects)

            if self.just_changed_fullscreen:
                pygame.display.flip()
                self.just_changed_fullscreen = False

        return self.restart

    def on_keydown(self, event):
        if event.key == pygame.K_ESCAPE:
            sys.exit()
        elif event.key == pygame.K_r:
            self.restart = True
            self.done = True
        elif event.key == pygame.K_f:
            if not self.fullscreen:
                pygame.display.set_mode((const.WIDTH, const.HEIGHT), pygame.FULLSCREEN)
                self.just_changed_fullscreen = True
                self.fullscreen = True
            else:
                pygame.display.set_mode((const.WIDTH, const.HEIGHT))
                self.just_changed_fullscreen = True
                self.fullscreen = False

    def on_mousedown(self, event):
        real_pos = (event.pos[0] / 2, event.pos[1] / 2)
        self.board_mgr.reveal(util.get_tile_at_xy(*real_pos))


class ConnectedComponent(object):

    def __init__(self, tiles, side, number):
        self.tiles = tiles
        self.side = side
        self.number = number

    def add(self, tile, number):
        # assume neighbor
        self.tiles.append(tile)
        self.number += number

    def merge(self, cc):
        self.tiles += cc.tiles
        self.tiles = sorted(list(set(self.tiles)))
        self.number += cc.number

    def get_center_tile(self):
        xs = [tile[0] for tile in self.tiles]
        ys = [tile[1] for tile in self.tiles]
        min_x = min(xs)
        max_x = max(xs)
        min_y = min(ys)
        max_y = max(ys)
        center = ((max_x + min_x) / 2, (max_y + min_y) / 2)
        return min([(math.hypot(tile[0] - center[0], tile[1] - center[1]), tile) for tile in self.tiles])[1]

    def __repr__(self):
        return "<" + str(self.tiles) + " " + self.side + " " + str(self.number) + ">"


class SlashManager(pygame.sprite.RenderUpdates):

    def __init__(self):
        pygame.sprite.RenderUpdates.__init__(self)
        self.image = pygame.image.load(data.filepath('slash.png'))
        self.timer = const.TIME_TO_DISAPPEAR

    def add_slash(self, critter):
        sprite = pygame.sprite.Sprite()
        sprite.image = self.image
        sprite.rect = sprite.image.get_rect()
        sprite.rect.center = critter.rect.center
        sprite.timer = self.timer
        sprite.critter = critter
        self.add(sprite)

    def update(self, dt):
        for sprite in self:
            sprite.timer -= dt
            if sprite.timer <= 0:
                self.remove(sprite)
                sprite.critter.done = True


class CritterManager(pygame.sprite.RenderUpdates):

    def __init__(self, slash_mgr):
        pygame.sprite.RenderUpdates.__init__(self)
        self.hidden_images = [pygame.image.load(data.filepath('critter1_hidden.png')).convert_alpha()]
        self.good_images = [pygame.image.load(data.filepath('critter1_good.png')).convert_alpha()]
        self.bad_images = [pygame.image.load(data.filepath('critter1_bad.png')).convert_alpha()]
        self.tiles = [[[] for j in range(const.NTILES_X)] for i in range(const.NTILES_Y)]
        self.slash_mgr = slash_mgr

    def add_hidden(self, c, i, j, offset):
        sprite = pygame.sprite.Sprite()
        sprite.image = self.hidden_images[c]
        sprite.rect = sprite.image.get_rect()
        pos = util.get_xy_at_tile(i, j)
        sprite.rect.center = [pos[0] + offset[0], pos[1] + offset[1]]
        sprite.imgtype = c
        sprite.tile = (i, j)
        sprite.done = False
        self.add(sprite)
        self.tiles[i][j].append(sprite)

    def set_side(self, i, j, side):
        for spr in self.tiles[i][j]:
            if side == 'n':
                spr.image = self.hidden_images[spr.imgtype]
            elif side == 'p':
                spr.image = self.good_images[spr.imgtype]
            elif side == 'e':
                spr.image = self.bad_images[spr.imgtype]

    def kill(self, critter):
        self.slash_mgr.add_slash(critter)

    def update(self, dt):
        for critter in self:
            if critter.done:
                i, j = critter.tile
                self.tiles[i][j].remove(critter)
                self.remove(critter)


class TileManager(pygame.sprite.RenderUpdates):

    def __init__(self):
        pygame.sprite.RenderUpdates.__init__(self)
        self.hidden_image = pygame.image.load(data.filepath('tile_hidden.png'))
        self.neutral_image = pygame.image.load(data.filepath('tile_neutral.png'))
        self.good_image = pygame.image.load(data.filepath('tile_good.png'))
        self.good_image_flipped = pygame.transform.flip(self.good_image, True, False)
        self.good_image_win = pygame.image.load(data.filepath('tile_good_win.png'))
        self.good_image_win_flipped = pygame.transform.flip(self.good_image_win, True, False)
        self.bad_image = pygame.image.load(data.filepath('tile_bad.png'))
        self.bad_image_flipped = pygame.transform.flip(self.bad_image, True, False)
        self.tiles = [[None for j in range(const.NTILES_X)] for i in range(const.NTILES_Y)]
        self.add_tiles()

    def add_tiles(self):
        for i in range(const.NTILES_Y):
            for j in range(const.NTILES_X):
                self.tiles[i][j] = self.add_tile(util.get_xy_at_tile(i, j))

    def add_tile(self, pos):
        sprite = pygame.sprite.Sprite()
        sprite.image = self.hidden_image
        sprite.rect = sprite.image.get_rect()
        sprite.rect.topleft = pos
        self.add(sprite)
        return sprite

    def set_tile(self, i, j, side):
        if side == 'n':
            self.tiles[i][j].image = self.neutral_image
        elif side == 'p':
            if (i + j) % 2 == 0:
                self.tiles[i][j].image = self.good_image
            else:
                self.tiles[i][j].image = self.good_image_flipped
        elif side == 'e':
            if (i + j) % 2 == 0:
                self.tiles[i][j].image = self.bad_image
            else:
                self.tiles[i][j].image = self.bad_image_flipped

    def set_win(self):
        for i in range(const.NTILES_Y):
            for j in range(const.NTILES_X):
                if (i + j) % 2 == 0:
                    self.tiles[i][j].image = self.good_image_win
                else:
                    self.tiles[i][j].image = self.good_image_win_flipped


class BoardManager(pygame.sprite.RenderUpdates):

    def __init__(self, critter_mgr, tile_mgr):
        pygame.sprite.RenderUpdates.__init__(self)
        self.original_board = [[None for j in range(const.NTILES_X)] for i in range(const.NTILES_Y)]
        self.current_board = [[None for j in range(const.NTILES_X)] for i in range(const.NTILES_Y)]
        self.board_sprites = [[pygame.sprite.Sprite() for j in range(const.NTILES_X)] for i in range(const.NTILES_Y)]
        self.connected_components = []
        for i in range(const.NTILES_Y):
            for j in range(const.NTILES_X):
                self.add(self.board_sprites[i][j])
        self.max_val = 3
        self.factor = 0.45
        # self.unbalance_limit = [1.0, 1.2]
        self.unbalance_limit = [1, 7]
        self.n_tiles_player = int(const.NTILES_TOTAL * self.factor)
        self.n_tiles_enemy = const.NTILES_TOTAL - self.n_tiles_player
        self.colors = {'s' : (255, 255, 255), 'p': (255, 255, 255), 'e': (255, 255, 255), 'n': (0, 0, 0)}
        self.update_timer = 0
        self.reveal_timer = 0
        self.critter_mgr = critter_mgr
        self.tile_mgr = tile_mgr
        self.busy = False
        self.tilepos_revealed = None
        self.overpop_limit = 7
        self.count = 0

    def setup(self):
        player_tiles = [random.randint(1, self.max_val) for i in range(self.n_tiles_player)]
        enemy_tiles = [random.randint(1, self.max_val) for i in range(self.n_tiles_enemy)]
        # while (sum(enemy_tiles) <= self.unbalance_limit[0] * sum(player_tiles) or 
        #     sum(enemy_tiles) >= self.unbalance_limit[1] * sum(player_tiles)): # just to avoid unbalanced cases
        while (sum(enemy_tiles) <= sum(player_tiles) + self.unbalance_limit[0] or 
            sum(enemy_tiles) >= sum(player_tiles) + self.unbalance_limit[1]): # just to avoid unbalanced cases
            player_tiles = [random.randint(1, self.max_val) for i in range(self.n_tiles_player)]
            enemy_tiles = [random.randint(1, self.max_val) for i in range(self.n_tiles_enemy)]
        all_tiles = [('p', i) for i in player_tiles] + [('e', i) for i in enemy_tiles]
        random.shuffle(all_tiles)
        # print sum(player_tiles), sum(enemy_tiles)

        for i in range(const.NTILES_Y):
            for j in range(const.NTILES_X):
                self.original_board[i][j] = all_tiles.pop()

        for i in range(const.NTILES_Y):
            for j in range(const.NTILES_X):
                self.clear_number(i, j)
                number = self.original_board[i][j][1]
                offsets = random.choice(const.OFFSETS[number-1])
                for num in range(number):
                    self.critter_mgr.add_hidden(0, i, j, offsets[num])

    def reveal(self, tilepos):
        i, j = tilepos
        if self.busy or self.current_board[i][j] != None:
            return
        side, number = self.original_board[i][j]
        self.tile_mgr.set_tile(i, j, side)
        self.critter_mgr.set_side(i, j, side)
        self.draw_number(i, j, self.original_board[i][j][1], self.colors[self.original_board[i][j][0]])
        self.reveal_timer = const.TIME_TO_REVEAL
        self.tilepos_revealed = tilepos
        self.busy = True
        sound = pygame.mixer.Sound(data.filepath("reveal.ogg"))
        sound.set_volume(const.SOUND_VOLUME)
        sound.play()

    def reveal_continue(self):
        i, j = self.tilepos_revealed
        side, number = self.original_board[i][j]
        self.update_connected_components(side, self.tilepos_revealed, number)
        self.consolidate_connected_components()
        self.tilepos_revealed = None
        self.count += 1

    def update_connected_components(self, side, newtile, number):
        found = False
        to_battle = []
        for cc in self.connected_components:
            for tile in cc.tiles:
                if util.is_tile_neighbor(newtile, tile):
                    if cc.side == side:
                        cc.add(newtile, number if not found else 0) # add number only to the first one
                        found = True
                        break # move to next cc
                    else:
                        if cc not in to_battle:
                            to_battle.append(cc)
        if not found:
            self.connected_components.append(ConnectedComponent([newtile], side, number))
        joining_ccs = [cc for cc in self.connected_components if newtile in cc.tiles]
        main_cc = self.merge(joining_ccs)
        if main_cc not in to_battle:
            to_battle.append(main_cc)
        # print 'pre-battle:', self.connected_components
        if len(to_battle) > 1:
            player_number, enemy_number = self.battle(to_battle)
            self.kill_critters(to_battle, 'p', player_number)
            self.kill_critters(to_battle, 'e', enemy_number)

            main_battle_cc = self.merge(to_battle)
            if player_number == 0 and enemy_number > 0:
                main_battle_cc.side = 'e'
                main_battle_cc.number = enemy_number
            elif player_number > 0 and enemy_number == 0:
                main_battle_cc.side = 'p'
                main_battle_cc.number = player_number
            else:
                main_battle_cc.side = 'n'
                main_battle_cc.number = 0

            # print 'post-battle:', self.connected_components

            self.update_timer = const.TIME_TO_DISAPPEAR
        else:
            self.update_regions()

    def merge(self, joining_ccs):
        main_cc = joining_ccs[0]
        for cc in joining_ccs:
            if cc != main_cc:
                main_cc.merge(cc)
                if cc in self.connected_components:
                    self.connected_components.remove(cc)
        return main_cc

    def battle(self, to_battle):
        enemy_number = 0
        player_number = 0
        for cc in to_battle:
            if cc.side == 'p':
                player_number += cc.number
            else:
                enemy_number += cc.number
        # print 'battle', player_number, enemy_number
        if player_number > enemy_number:
            player_number -= (enemy_number + 1) / 2
            enemy_number = 0
        elif player_number < enemy_number:
            enemy_number -= (player_number + 1) / 2
            player_number = 0
        else:
            player_number = 0
            enemy_number = 0
        # print 'result:', player_number, enemy_number
        return player_number, enemy_number

    def kill_critters(self, to_kill, side, updated_number):
        killed = False
        critters = []
        for cc in to_kill:
            for i, j in cc.tiles:
                if side == None or cc.side == side:
                    critters += self.critter_mgr.tiles[i][j]
        if len(critters) > updated_number:
            random.shuffle(critters)
            for k in range(len(critters) - updated_number):
                self.critter_mgr.kill(critters[k])
                killed = True
        if killed:
            sound = pygame.mixer.Sound(data.filepath("hit.ogg"))
            sound.set_volume(const.SOUND_VOLUME)
            sound.play()

    def consolidate_connected_components(self):
        self.current_board = [[None for j in range(const.NTILES_X)] for i in range(const.NTILES_Y)]
        for cc in self.connected_components:
            for tile in cc.tiles:
                assert(self.current_board[tile[0]][tile[1]] == None)
                self.current_board[tile[0]][tile[1]] = cc

    def check_overpopulation(self):
        overpopulation = False
        for cc in self.connected_components:
            if cc.number > self.overpop_limit:
                cc.number = self.overpop_limit
                overpopulation = True
                self.kill_critters([cc], None, self.overpop_limit)
        self.update_timer = const.TIME_TO_DISAPPEAR
        return overpopulation

    def update_regions(self):
        for cc in self.connected_components:
            for i, j in cc.tiles:
                self.tile_mgr.set_tile(i, j, cc.side)
                self.critter_mgr.set_side(i, j, cc.side)
                self.clear_number(i, j)
            center_tile = cc.get_center_tile()
            self.draw_number(center_tile[0], center_tile[1], cc.number, self.colors[cc.side])
        if not self.check_overpopulation():
            self.busy = False

    def draw_number(self, i, j, number, color):
        textpos = list(util.get_xy_at_tile_center(i, j))
        textpos[1] -= const.TILE_SIZE * 0.4
        extra_str = ""
        if number > self.overpop_limit:
            extra_str = "\nover " + str(self.overpop_limit)
        if self.count == const.NTILES_TOTAL:
            if self.connected_components[0].side == 'p':
                extra_str = "\n\nYou win!"
                self.tile_mgr.set_win()
            elif self.connected_components[0].side == 'n':
                extra_str = "\n\nIt's a tie!"
            else:
                extra_str = "\n\nThe enemy wins"
        self.board_sprites[i][j].image, pos = ptext.draw(str(number) + extra_str, textpos, color=color, fontsize=16,
            fontname=const.FONT_MAIN, align="center")
        self.board_sprites[i][j].rect = self.board_sprites[i][j].image.get_rect()
        self.board_sprites[i][j].rect.midtop = pos

    def clear_number(self, i, j):
        self.board_sprites[i][j].image = pygame.surface.Surface((0, 0))
        self.board_sprites[i][j].rect = self.board_sprites[i][j].image.get_rect()

    def update(self, dt):
        if self.update_timer > 0:
            self.update_timer -= dt
            if self.update_timer <= 0:
                self.update_timer = 0
                self.update_regions()
        if self.reveal_timer > 0:
            self.reveal_timer -= dt
            if self.reveal_timer <= 0:
                self.reveal_timer = 0
                self.reveal_continue()


if __name__ == '__main__':
    pygame.init()
    GameWindow()
